import copy
import pickle

from polaris.params import Domain
from polaris.rabbitmq import JobClient


class Polaris(object):
    """
    A client for the Polaris.
    """

    def __init__(
            self, fn, bounds, algo, trials,
            max_evals=10, exp_key=None,
            logger=None, debug=False, args=None):
        """
        Parameters
        ----------
        fn : callable
            fnction which Polaris will evaluate
        bounds : array of params
            search range of hyperparameters
        algo : str
            search algorithms (random, tpe, bayesian)
        trials : polaris.trials
            trials object which store all parameters of trials
        max_evals : int
            the maximum number of evaluations
        exp_key : str
            the expriment key to distinguish each experiments
        logger : Logger
            user logger object
        debug : bool
            make Polaris debug mode
        args : tuple
            args to pass fn
        """

        self.fn = fn
        self.bounds = bounds
        self.algo = algo
        self.trials = trials
        self.max_evals = max_evals
        self.debug = debug
        self.args = args
        if exp_key is None:
            self.exp_key = fn.__name__
        else:
            self.exp_key = exp_key
        self.exp_info = {
            'exp_key': exp_key,
            'eval_count': len(trials) + 1,
        }

        if logger is None:
            import logging
            self.logger = logging.getLogger(__name__)

            if self.debug:
                self.logger.setLevel(logging.DEBUG)
            else:
                self.logger.setLevel(logging.INFO)
            stream = logging.StreamHandler()
            formatter = logging.Formatter(
                    '%(asctime)s:%(lineno)d:%(levelname)s:%(message)s')
            stream.setFormatter(formatter)
            self.logger.addHandler(stream)
        else:
            self.logger = logger

        self.domain = Domain(self.bounds, algo=self.algo)

    def run(self, use_mpi=False):
        """
        Run an experiment sequentially with one process up to max_evals count.
        The expriment will early stop, if params become None

        Parameters
        ----------
        use_mpi : boolean
            boolean for using mpi or not
        """

        if use_mpi:
            self.run_with_mpi()
        else:
            self.run_single()

    def run_single(self):
        self.logger.info('Start searching...')

        for eval_count in range(self.exp_info['eval_count'], self.max_evals+1):
            params = self.domain.predict(self.trials)[0]
            fn_params = copy.copy(params)

            if self.args is None:
                exp_result = self.fn(fn_params, self.exp_info)
            else:
                exp_result = self.fn(fn_params, self.exp_info, *self.args)

            self.trials.add(exp_result, params, self.exp_info)
            self.exp_info['eval_count'] += 1
            self.logger.debug(fn_params)

            with open(f'{self.exp_key}.p', mode='wb') as f:
                pickle.dump(self.trials, f)

        return self.trials.best_params

    def run_with_mpi(self):
        from mpi4py import MPI
        comm = MPI.COMM_WORLD
        rank = comm.Get_rank()

        for eval_count in range(self.exp_info['eval_count'], self.max_evals+1):
            if rank == 0:
                params = self.domain.predict(self.trials)[0]
                fn_params = copy.copy(params)
            else:
                fn_params = None
            fn_params = comm.bcast(fn_params, root=0)

            if self.args is None:
                exp_result = self.fn(fn_params, self.exp_info)
            else:
                exp_result = self.fn(fn_params, self.exp_info, *self.args)

            if rank == 0:
                self.trials.add(exp_result, params, self.exp_info)
                self.exp_info['eval_count'] += 1
                self.logger.debug(fn_params)

                with open(f'{self.exp_key}.p', mode='wb') as f:
                    pickle.dump(self.trials, f)

        if rank == 0:
            return self.trials.best_params

    def run_parallel(self):
        """
        Run an experiment in parallel.

        Polaris use RabbitMQ to pass arguments from client to worker.
        You need to start at least one worker to start an experiment.
        """

        job_client = JobClient(self)
        job_client.start()
        return self.trials.best_params



from polaris import STATUS_SUCCESS


def pseudo_train(params, exp_info):
    lr_squared = ((params['lr'] - 0.006) ** 2) * 1e+5
    weight_decay_squared = ((params['weight_decay'] - 0.02) ** 2) * 1e+5
    loss = lr_squared + weight_decay_squared
    return {
        'loss':  loss,
        'status': STATUS_SUCCESS,
    }



from functools import partial
import random

import chainer
import chainer.links as L
from chainer import training
from chainer.training import extensions
from chainer.datasets import TransformDataset, get_cifar10
from chainercv import transforms
import chainermn

import cv2 as cv
from mpi4py import MPI
import numpy as np
from polaris import STATUS_SUCCESS, STATUS_FAILURE

import models.resnet50
import models.lenet5

OPTIMIZERS = {
    'momentum_sgd': chainer.optimizers.MomentumSGD,
    'adam': chainer.optimizers.Adam,
    }

ARCHS = {
    'lenet5': models.lenet5.LeNet5,
    'resnet50': models.resnet50.ResNet50,
}


OMIT_ARG_PARAMS = [
    'lr_decay_rate',
    'lr_decay_epoch',
    'weight_decay',
]

mpi_comm = MPI.COMM_WORLD


def cv_rotate(img, angle):
    img = img.transpose(1, 2, 0) / 255.
    center = (img.shape[0] // 2, img.shape[1] // 2)
    r = cv.getRotationMatrix2D(center, angle, 1.0)
    img = cv.warpAffine(img, r, img.shape[:2])
    img = img.transpose(2, 0, 1) * 255.
    img = img.astype(np.float32)
    return img


def transform(
        inputs, mean, std, random_angle=15., pca_sigma=255., expand_ratio=1.0,
        crop_size=(32, 32), train=True):
    img, label = inputs
    img = img.copy()

    # Random rotate
    if random_angle != 0:
        angle = np.random.uniform(-random_angle, random_angle)
        img = cv_rotate(img, angle)

    # Color augmentation
    if train and pca_sigma != 0:
        img = transforms.pca_lighting(img, pca_sigma)

    # Standardization
    img -= mean[:, None, None]
    img /= std[:, None, None]

    if train:
        # Random flip
        img = transforms.random_flip(img, x_random=True)
        # Random expand
        if expand_ratio > 1:
            img = transforms.random_expand(img, max_ratio=expand_ratio)
        # Random crop
        if tuple(crop_size) != (32, 32):
            img = transforms.random_crop(img, tuple(crop_size))

    return img, label


def run(params, exp_info, args):
    random.seed(args.seed)
    np.random.seed(args.seed)
    chainer.cuda.cupy.random.seed(args.seed)

    # Prepare ChainerMN communicator.
    if args.communicator == 'naive':
        print("Error: 'naive' communicator does not support GPU.\n")
        exit(-1)

    comm = chainermn.create_communicator(args.communicator, mpi_comm)
    device = comm.intra_rank

    if comm.rank == 0:
        print('==========================================')
        print('Experiment Info: {}'.format(exp_info))
        print('Num process (COMM_WORLD): {}'.format(comm.size))
        print('Using GPUs')
        print('Using {} communicator'.format(args.communicator))
        print('Num Minibatch-size: {}'.format(args.batchsize))
        print('Num epoch: {}'.format(args.epoch))
        print('==========================================')

    # Set up a neural network to train.
    # Classifier reports softmax cross entropy loss and accuracy at every
    # iteration, which will be used by the PrintReport extension below.
    class_labels = 10
    model = L.Classifier(ARCHS[args.arch](class_labels))
    if device >= 0:
        chainer.cuda.get_device_from_id(device).use()
        model.to_gpu()

    arg_params = {i: params[i] for i in params if i not in OMIT_ARG_PARAMS}
    optimizer = chainermn.create_multi_node_optimizer(
            OPTIMIZERS[args.optimizer_name](**arg_params), comm)
    optimizer.setup(model)
    optimizer.add_hook(
            chainer.optimizer_hooks.WeightDecay(params['weight_decay']))

    if comm.rank == 0:
        train, valid = get_cifar10(scale=255.)
        mean = np.mean([x for x, _ in train], axis=(0, 2, 3))
        std = np.std([x for x, _ in train], axis=(0, 2, 3))
    else:
        train, valid, mean, std = None, None, None, None
    train = chainermn.scatter_dataset(train, comm, shuffle=True)
    valid = chainermn.scatter_dataset(valid, comm, shuffle=True)
    mean = mpi_comm.bcast(mean, root=0)
    std = mpi_comm.bcast(std, root=0)

    train_transform = partial(
        transform, mean=mean, std=std, random_angle=args.random_angle,
        pca_sigma=args.pca_sigma, expand_ratio=args.expand_ratio,
        crop_size=args.crop_size, train=True)
    valid_transform = partial(transform, mean=mean, std=std, train=False)
    train = TransformDataset(train, train_transform)
    valid = TransformDataset(valid, valid_transform)

    train_iter = chainer.iterators.SerialIterator(train, args.batchsize)
    test_iter = chainer.iterators.SerialIterator(valid, args.batchsize,
                                                 repeat=False, shuffle=True)

    updater = training.StandardUpdater(train_iter, optimizer, device=device)
    stop_trigger = (args.epoch, 'epoch')
    trainer = training.Trainer(updater, stop_trigger, out=args.out)
    # Create a multi node evaluator from a standard Chainer evaluator.
    evaluator = extensions.Evaluator(test_iter, model, device=device)
    evaluator = chainermn.create_multi_node_evaluator(evaluator, comm)
    trainer.extend(evaluator)

    if comm.rank == 0:
        # Some display and output extensions are necessary only for one worker.
        # (Otherwise, there would just be repeated outputs.)
        trainer.extend(extensions.dump_graph('main/loss'))
        trainer.extend(extensions.LogReport())
        trainer.extend(extensions.PrintReport(
            ['epoch', 'main/loss', 'validation/main/loss',
             'main/accuracy', 'validation/main/accuracy', 'elapsed_time']))
        trainer.extend(extensions.ProgressBar())

    if args.optimizer_name != 'adam':
        trainer.extend(extensions.ExponentialShift(
            'lr', params['lr_decay_rate']),
            trigger=(params['lr_decay_epoch'], 'epoch'))

    # Run the training
    trainer.run()

    if comm.rank == 0:
        print('========== Done ===========')

        loss = trainer.observation['validation/main/loss']
        loss = chainer.cuda.to_cpu(loss)

        if np.isnan(loss):
            return {
                'status': STATUS_FAILURE
            }
        else:
            return {
                'loss': loss,
                'status': STATUS_SUCCESS
            }



import warnings

from .polaris import Polaris
from .params import Bounds, Domain
from .trials import Trials, STATUS_SUCCESS, STATUS_FAILURE

warnings.simplefilter("ignore", UserWarning)

__all__ = [
    'Bounds',
    'Domain',
    'Polaris',
    'Trials',
    'STATUS_SUCCESS',
    'STATUS_FAILURE'
]



import numpy as np
import sklearn.gaussian_process as gp
from scipy.optimize import minimize
from scipy.stats import norm


def expected_improvement(x, model, lowest_loss):
    next_params = x.reshape(1, -1)
    mu, sigma = model.predict(next_params, return_std=True)

    with np.errstate(divide='ignore'):
        diff = lowest_loss - mu
        z = diff / sigma
        ei = (lowest_loss - mu) * norm.cdf(z) + sigma * norm.pdf(z)
        ei[sigma == 0.0] == 0.0

    return -1 * ei


def calc_next_params(domain, trials):
    # At first time, get random params
    if len(trials) <= 5:
        next_params = {}
        random_result = domain.random()
        for index, fieldname in enumerate(domain.fieldnames):
            next_params[fieldname] = random_result[index]
        return [next_params]

    train_x, train_y = trials.get_train_data()

    kernel = gp.kernels.Matern()
    model = gp.GaussianProcessRegressor(
            kernel=kernel,
            n_restarts_optimizer=9)

    model.fit(train_x, train_y)

    lowest_loss = trials.lowest_loss

    best_ei = -np.Inf
    n_iter = 400
    best_xs = []

    for _ in range(0, n_iter):
        minimize_result = minimize(
                fun=expected_improvement,
                x0=domain.random(),
                bounds=domain.bounds,
                method='L-BFGS-B',
                args=(model, lowest_loss),
                )
        ei = -minimize_result.fun
        if ei > best_ei:
            best_ei = ei
            best_xs.insert(0, minimize_result.x)

    next_params_candidates = []
    for best_x in best_xs:
        next_params = {}
        for index, fieldname in enumerate(domain.fieldnames):
            next_params[fieldname] = best_x[index]
        next_params_candidates.append(next_params)

    return next_params_candidates



import pickle

import pika

from polaris.rabbitmq.config import (
    RABBITMQ_HOST,
    RABBITMQ_PORT,
    RABBITMQ_VIRTUAL_HOST,
    RABBITMQ_USERNAME,
    RABBITMQ_PASSWORD,
)


class JobClient(object):
    """
    A client class for parallel experiments.

    The instance of this class send a new job to workers and
    calculate next parameters in response to the requsest from a worker.

    This client adopt the RPC pattern.
    Therefore all results will be accumulated on client side.
    """

    def __init__(self, polaris):
        self.parallel_count = 0
        self.next_params_candidates = []

        self.polaris = polaris
        self.exp_key = polaris.exp_key
        self.logger = polaris.logger

        self.job_queue_name = f'job_{self.exp_key}'
        self.request_queue_name = f'request_{self.exp_key}'

        if RABBITMQ_USERNAME and RABBITMQ_PASSWORD:
            credentials = pika.PlainCredentials(
                    RABBITMQ_USERNAME, RABBITMQ_PASSWORD)
            rabbitmq_params = pika.ConnectionParameters(
                    host=RABBITMQ_HOST,
                    port=RABBITMQ_PORT,
                    virtual_host=RABBITMQ_VIRTUAL_HOST,
                    credentials=credentials)
        else:
            rabbitmq_params = pika.ConnectionParameters(
                    host=RABBITMQ_HOST,
                    port=RABBITMQ_PORT,
                    virtual_host=RABBITMQ_VIRTUAL_HOST)

        self.connection = pika.BlockingConnection(rabbitmq_params)

        self.channel = self.connection.channel()
        result = self.channel.queue_declare(exclusive=True)
        self.channel.queue_declare(queue=self.request_queue_name)
        self.channel.queue_declare(queue=self.job_queue_name)
        self.channel.basic_qos(prefetch_count=1)

        self.callback_queue = result.method.queue
        self.channel.basic_consume(
                self.on_response, no_ack=True, queue=self.callback_queue)
        self.channel.basic_consume(
                self.on_request, no_ack=True, queue=self.request_queue_name)

    def on_request(self, ch, method, props, body):
        """
        A method to receive job request from workers
        After receiving request, this method will send a job to them.
        """

        self.send_job()

    def send_job(self):
        domain = self.polaris.domain
        trials = self.polaris.trials

        eval_count = self.polaris.exp_info['eval_count']
        max_evals = self.polaris.max_evals
        if eval_count > max_evals:
            return

        if self.parallel_count == 0:
            self.next_params_candidates = domain.predict(trials)
            next_params = self.next_params_candidates[0]
        else:
            candidates_count = len(self.next_params_candidates)
            if self.parallel_count < candidates_count:
                next_params = self.next_params_candidates[self.parallel_count]
            else:
                next_params = {}
                random_params = domain.random()
                for index, fieldname in enumerate(domain.fieldnames):
                    next_params[fieldname] = random_params[index]
        self.parallel_count += 1

        fn = self.polaris.fn
        fn_module = fn.__module__
        fn_name = fn.__name__
        if fn_module == '__main__':
            # TODO This transformation is ugly...
            import __main__
            fn_module = __main__.__file__.replace('/', '.').replace('.py', '')

        ctx = {
            'fn_module': fn_module,
            'fn_name': fn_name,
            'params': next_params,
            'exp_info': self.polaris.exp_info,
            'args': self.polaris.args,
        }

        self.channel.basic_publish(
                exchange='',
                routing_key=self.job_queue_name,
                properties=pika.BasicProperties(
                    reply_to=self.callback_queue,
                    ),
                body=pickle.dumps(ctx)
                )

        self.polaris.exp_info['eval_count'] += 1

    def on_response(self, ch, method, props, body):
        """
        A method to receive the result of experiment from workers.
        """
        exp_payload = pickle.loads(body)
        exp_result = exp_payload['exp_result']
        params = exp_payload['params']
        exp_info = exp_payload['exp_info']

        self.polaris.trials.add(exp_result, params, exp_info)
        self.parallel_count = 0

        with open(f'{self.exp_key}.p', mode='wb') as f:
            pickle.dump(self.polaris.trials, f)

        eval_count = len(self.polaris.trials)
        max_evals = self.polaris.max_evals

        if eval_count >= max_evals:
            self.connection.close()

    def start(self):
        """
        A method to start consuming job requests.
        """

        self.logger.info('Start parallel execution...')

        try:
            self.channel.start_consuming()
        except (pika.exceptions.ChannelClosed, KeyboardInterrupt):
            self.channel.queue_delete(queue=self.request_queue_name)
            self.channel.queue_delete(queue=self.job_queue_name)
            self.logger.info('All jobs have finished')



import chainer
import chainer.functions as F
import chainer.links as L


class BottleNeck(chainer.Chain):

    def __init__(self, n_in, n_mid, n_out, stride=1, use_conv=False):
        w = chainer.initializers.HeNormal()
        super(BottleNeck, self).__init__()
        with self.init_scope():
            self.conv1 = L.Convolution2D(n_in, n_mid, 1, stride, 0, True, w)
            self.bn1 = L.BatchNormalization(n_mid)
            self.conv2 = L.Convolution2D(n_mid, n_mid, 3, 1, 1, True, w)
            self.bn2 = L.BatchNormalization(n_mid)
            self.conv3 = L.Convolution2D(n_mid, n_out, 1, 1, 0, True, w)
            self.bn3 = L.BatchNormalization(n_out)
            if use_conv:
                self.conv4 = L.Convolution2D(
                    n_in, n_out, 1, stride, 0, True, w)
                self.bn4 = L.BatchNormalization(n_out)
        self.use_conv = use_conv

    def __call__(self, x):
        h = F.relu(self.bn1(self.conv1(x)))
        h = F.relu(self.bn2(self.conv2(h)))
        h = self.bn3(self.conv3(h))
        return h + self.bn4(self.conv4(x)) if self.use_conv else h + x


class Block(chainer.ChainList):

    def __init__(self, n_in, n_mid, n_out, n_bottlenecks, stride=2):
        super(Block, self).__init__()
        self.add_link(BottleNeck(n_in, n_mid, n_out, stride, True))
        for _ in range(n_bottlenecks - 1):
            self.add_link(BottleNeck(n_out, n_mid, n_out))

    def __call__(self, x):
        for f in self:
            x = f(x)
        return x


class ResNet(chainer.Chain):

    def __init__(self, n_class=10, n_blocks=[3, 4, 6, 3]):
        super(ResNet, self).__init__()
        w = chainer.initializers.HeNormal()
        with self.init_scope():
            self.conv1 = L.Convolution2D(None, 64, 3, 1, 0, True, w)
            self.bn2 = L.BatchNormalization(64)
            self.res3 = Block(64, 64, 256, n_blocks[0], 1)
            self.res4 = Block(256, 128, 512, n_blocks[1], 2)
            self.res5 = Block(512, 256, 1024, n_blocks[2], 2)
            self.res6 = Block(1024, 512, 2048, n_blocks[3], 2)
            self.fc7 = L.Linear(None, n_class)

    def __call__(self, x):
        h = F.relu(self.bn2(self.conv1(x)))
        h = self.res3(h)
        h = self.res4(h)
        h = self.res5(h)
        h = self.res6(h)
        h = F.average_pooling_2d(h, h.shape[2:])
        h = self.fc7(h)
        return h


class ResNet50(ResNet):

    def __init__(self, n_class=10):
        super(ResNet50, self).__init__(n_class, [3, 4, 6, 3])



import numpy as np


from polaris.optimizers import bayesian_opt, rand, tpe

OPTIMIZERS = {
    'bayesian_opt': bayesian_opt.calc_next_params,
    'random': rand.calc_next_params,
    'tpe': tpe.calc_next_params
}


class Bounds:
    """
    A class to declare range of each hyperparameters.
    For now, Polaris do not adapt to category variable.
    """

    def __init__(self, label, low, high, q=None):
        self.label = label
        self.low = low
        self.high = high
        self.q = q

    def range(self):
        return self.high - self.low


class Domain:
    """
    A class to store bounds and searching new parameters.
    """

    def __init__(self, bounds, algo='random'):
        self.n_params = len(bounds)
        self._algo = algo
        self._bounds = sorted(bounds, key=lambda b: b.label)

    @property
    def bounds(self):
        array_bounds = []
        for b in self._bounds:
            array_bounds.append((b.low, b.high))
        return array_bounds

    @property
    def fieldnames(self):
        labels = []
        for b in self._bounds:
            labels.append(b.label)
        return labels

    def random(self):
        rand_param = []
        r = np.random.rand(1, self.n_params)
        for i, b in enumerate(self._bounds):
            p = b.low + r[0][i] * b.range()
            if b.q is None:
                rand_param.append(p)
            else:
                rand_param.append(round(p / b.q) * b.q)
        return np.array(rand_param)

    def predict(self, trials):
        return OPTIMIZERS[self._algo](self, trials)



from setuptools import setup, find_packages


def get_long_description():
    with open('README.md') as f:
        readme = f.read()
    return readme


setup(
    name='polaris-py',
    packages=find_packages(),
    scripts=['bin/polaris-worker'],
    version='0.7.2',
    author='Shun Iwase, Linsho Kaku',
    author_email='s@sh8.io, kaku.l.aa@rio.gsic.titech.ac.jp',
    description='Polaris is a hyperparamter optimization library.',
    long_description=get_long_description(),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Environment :: Console',
        'License :: OSI Approved :: MIT License',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6'
        ],
    platforms=['Linux', 'OS-X'],
    license='MIT',
    install_requires=[
        'numpy',
        'scikit-learn',
        'scipy',
        'pika==0.12.0',
        'statsmodels',
        'matplotlib',
    ],
    tests_require=['pytest'],
)



import copy
import functools
import importlib
import pickle

import pika

from polaris.rabbitmq.config import (
    RABBITMQ_HOST,
    RABBITMQ_PORT,
    RABBITMQ_VIRTUAL_HOST,
    RABBITMQ_USERNAME,
    RABBITMQ_PASSWORD,
)


class JobWorker(object):
    """
    A worker class for parallel experiments.

    You can start the worker like below.
    `polaris-worker --exp-key this_is_test`

    And if you want to run this worker on multi node environment,
    you have to add `--mpi` flag.
    """

    def __init__(self, args, logger=None, debug=False):
        self.exp_key = args.exp_key
        self.use_mpi = args.mpi
        self.debug = debug

        if self.use_mpi:
            from mpi4py import MPI
            self.comm = MPI.COMM_WORLD
            self.rank = self.comm.Get_rank()

        if (not self.use_mpi) or self.rank == 0:
            self.job_queue_name = f'job_{self.exp_key}'
            self.request_queue_name = f'request_{self.exp_key}'

            if RABBITMQ_USERNAME and RABBITMQ_PASSWORD:
                credentials = pika.PlainCredentials(
                        RABBITMQ_USERNAME, RABBITMQ_PASSWORD)
                rabbitmq_params = pika.ConnectionParameters(
                        host=RABBITMQ_HOST,
                        port=RABBITMQ_PORT,
                        virtual_host=RABBITMQ_VIRTUAL_HOST,
                        credentials=credentials,
                        heartbeat_interval=0)
            else:
                rabbitmq_params = pika.ConnectionParameters(
                        host=RABBITMQ_HOST,
                        port=RABBITMQ_PORT,
                        virtual_host=RABBITMQ_VIRTUAL_HOST,
                        heartbeat_interval=0)

            self.connection = pika.BlockingConnection(rabbitmq_params)
            self.channel = self.connection.channel()
            self.channel.queue_declare(queue=self.request_queue_name)
            self.channel.queue_declare(queue=self.job_queue_name)
            self.channel.basic_qos(prefetch_count=1)
            self.channel.basic_consume(
                    self.on_request, queue=self.job_queue_name)

        if logger is None:
            import logging
            self.logger = logging.getLogger(__name__)

            if self.debug:
                self.logger.setLevel(logging.DEBUG)
            else:
                self.logger.setLevel(logging.INFO)

            stream = logging.StreamHandler()
            formatter = logging.Formatter(
                    '%(asctime)s:%(lineno)d:%(levelname)s:%(message)s')
            stream.setFormatter(formatter)
            self.logger.addHandler(stream)
        else:
            self.logger = logger

    def start(self):
        try:
            if (not self.use_mpi) or self.rank == 0:
                self.logger.info('Waiting for new job...')

                self.request_job()

                self.channel.start_consuming()
            else:
                while True:
                    ctx = None
                    ctx = self.comm.bcast(ctx, root=0)
                    self.run(ctx)
        except KeyboardInterrupt:
            self.logger.info('Stop current worker...')

            if (not self.use_mpi) or self.rank == 0:
                self.connection.close()

            if self.use_mpi:
                from mpi4py import MPI
                MPI.Finalize()

    def on_request(self, ch, method, props, body):
        ctx = pickle.loads(body)
        self.logger.info(ctx)

        if self.use_mpi:
            ctx = self.comm.bcast(ctx, root=0)

        exp_result = self.run(ctx)

        exp_payload = {
            'exp_result': exp_result,
            'params': ctx['params'],
            'exp_info': ctx['exp_info'],
        }

        self.channel.basic_publish(
            exchange='',
            routing_key=props.reply_to,
            body=pickle.dumps(exp_payload)
        )

        self.request_job()

        self.channel.basic_ack(delivery_tag=method.delivery_tag)

    def request_job(self):
        self.channel.basic_publish(
                exchange='',
                routing_key=self.request_queue_name,
                body=''
            )

    def run(self, ctx):
        params = ctx['params']
        exp_info = ctx['exp_info']
        fn_module = ctx['fn_module']
        fn_name = ctx['fn_name']
        args = ctx['args']

        fn_params = copy.copy(params)
        fn_module = importlib.import_module(fn_module)
        fn = getattr(fn_module, fn_name)

        if args is not None:
            exp_result = fn(fn_params, exp_info, *args)
        else:
            exp_result = fn(fn_params, exp_info)

        return exp_result
